# SPDX-FileCopyrightText: Copyright (c) 2024 Refeyn Ltd and other h5view contributors
# SPDX-License-Identifier: MIT

if __name__ == "__main__":
    try:
        from . import main
    except ImportError:
        from h5view import main

    main()



# SPDX-FileCopyrightText: Copyright (c) 2024 Refeyn Ltd and other h5view contributors
# SPDX-License-Identifier: MIT

import math
import random
from typing import Any, cast, List, Optional, Tuple, Union

import h5py
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from h5view import utils


class DatasetModel(QtCore.QAbstractTableModel):
    def __init__(self, dataset: h5py.Dataset):
        super().__init__()
        self._dataset = dataset
        self._rows = self._cols = 1
        if dataset.ndim == 1:
            self._rows = dataset.size
        elif dataset.ndim > 1:
            self._rows, self._cols = dataset.shape[-2:]
        self._region = (0,) * len(dataset.shape[:-2])

    def rowCount(
        self,
        parent: Union[
            QtCore.QModelIndex, QtCore.QPersistentModelIndex
        ] = QtCore.QModelIndex(),
    ) -> int:
        if parent.isValid():
            return 0
        return self._rows

    def columnCount(
        self,
        parent: Union[
            QtCore.QModelIndex, QtCore.QPersistentModelIndex
        ] = QtCore.QModelIndex(),
    ) -> int:
        if parent.isValid():
            return 0
        return self._cols

    def region(self) -> Tuple[int, ...]:
        return self._region

    def regionBounds(self) -> Tuple[int, ...]:
        return tuple(self._dataset.shape[:-2])

    def setRegion(self, region: Tuple[int, ...]) -> None:
        assert len(region) == len(self._region)
        self._region = region
        self.dataChanged.emit(
            self.index(0, 0), self.index(self._rows - 1, self._cols - 1), []
        )

    def data(
        self,
        index: Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex],
        role: int = cast(int, QtCore.Qt.ItemDataRole.DisplayRole),
    ) -> Any:
        if not index.isValid() or any(s == 0 for s in self._dataset.shape):
            return None
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            didx = self._region + (index.row(), index.column())
            return utils.formatAsStr(self._dataset[didx[: self._dataset.ndim]])


class PixmapWidget(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        self._dataset: Optional[h5py.Dataset] = None
        self._rows: Optional[int] = None
        self._cols: Optional[int] = None
        self._region: Optional[Tuple[int, ...]] = None
        self._autoscale = False

    def setDataset(self, dataset: h5py.Dataset) -> None:
        self._dataset = dataset
        self._rows = self._cols = 1
        if dataset.ndim == 1:
            self._rows = dataset.size
        elif dataset.ndim > 1:
            self._cols, self._rows = dataset.shape[-2:]
        self._region = (0,) * len(dataset.shape[:-2])
        self.repaint()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        if self._dataset is None:
            return
        image = self.getImage()
        painter = QtGui.QPainter(self)
        imageRect = QtCore.QRect()
        imageRect.setSize(
            image.size().scaled(self.size(), QtCore.Qt.AspectRatioMode.KeepAspectRatio)
        )
        imageRect.moveCenter(QtCore.QPoint(self.width() // 2, self.height() // 2))
        painter.drawImage(imageRect, image)

    def region(self) -> Tuple[int, ...]:
        assert self._region is not None
        return self._region

    def regionBounds(self) -> Tuple[int, ...]:
        assert self._dataset is not None
        return tuple(self._dataset.shape[:-2])

    def setRegion(self, region: Tuple[int, ...]) -> None:
        assert self._region is not None
        assert len(region) == len(self._region)
        self._region = region
        self.repaint()

    def autoscale(self) -> bool:
        return self._autoscale

    def setAutoscale(self, autoscale: bool) -> None:
        self._autoscale = autoscale
        self.repaint()

    def getImage(self) -> QtGui.QImage:
        assert (
            self._rows is not None
            and self._cols is not None
            and self._region is not None
            and self._dataset is not None
        )
        if any(s == 0 for s in self._dataset.shape):
            return QtGui.QImage()
        data: np.ndarray = self._dataset[self._region].astype(float)
        if self._autoscale:
            data -= np.nanmin(data)
            data /= np.nanmax(data)
        else:
            dtypeInfo: Union[np.finfo, np.iinfo]
            if np.issubdtype(self._dataset.dtype, np.floating):
                dtypeInfo = np.finfo(self._dataset.dtype)
            else:
                dtypeInfo = np.iinfo(self._dataset.dtype)
            data /= dtypeInfo.max
            minByMax = dtypeInfo.min / dtypeInfo.max
            data -= minByMax
            data /= 1 - minByMax
        data = (data * 255).copy(order="C").astype(np.uint8)
        assert data.shape == (self._cols, self._rows)
        return QtGui.QImage(
            data.tobytes(),
            self._rows,
            self._cols,
            self._rows,
            QtGui.QImage.Format.Format_Grayscale8,
        )


def columnWidthRandomized(widget: QtWidgets.QTableView, column: int) -> float:
    model = widget.model()
    delegate = widget.itemDelegate()
    assert isinstance(delegate, QtWidgets.QStyledItemDelegate)
    widths: List[float] = []
    for _ in range(10):
        index = model.index(random.randrange(model.rowCount()), column)
        style = QtWidgets.QStyleOptionViewItem()
        delegate.initStyleOption(style, index)
        sh = delegate.sizeHint(style, index)
        widths.append(sh.width())
    return max(widths) + 12


def columnWidthFromDtype(
    widget: QtWidgets.QTableView, dtype: np.dtype
) -> Optional[float]:
    info: Union[np.finfo, np.iinfo]
    if np.issubdtype(dtype, np.floating):
        info = np.finfo(dtype)
        digits = math.ceil(info.bits * math.log10(2))
        # Add sign, decimal point, and exponent
        digits += 4
    elif np.issubdtype(dtype, np.integer):
        info = np.iinfo(dtype)
        digits = math.ceil(info.bits * math.log10(2))
        if np.issubdtype(dtype, np.signedinteger):
            # Add sign
            digits += 1
    else:
        return None
    model = widget.model()
    delegate = widget.itemDelegate()
    assert isinstance(delegate, QtWidgets.QStyledItemDelegate)
    index = model.index(0, 0)
    style = QtWidgets.QStyleOptionViewItem()
    delegate.initStyleOption(style, index)
    fontMetrics = QtGui.QFontMetrics(style.font)  # type: ignore[attr-defined]
    return float(fontMetrics.horizontalAdvance("8" * digits)) + 12



# SPDX-FileCopyrightText: Copyright (c) 2024 Refeyn Ltd and other h5view contributors
# SPDX-License-Identifier: MIT

from typing import Any, Mapping, Tuple, Union

import h5py
import numpy as np


def formatAsStr(value: Any) -> str:
    if isinstance(value, h5py.Group):
        return f"Group {value.name!r} with {len(value.keys())} members"
    elif isinstance(value, h5py.Dataset):
        if value.shape == ():
            singlevalue = value[()]
            if isinstance(singlevalue, bytes):
                try:
                    return singlevalue.decode()
                except ValueError:
                    pass
            return str(singlevalue)
        else:
            return (
                f"Dataset {value.name!r} {typeAndShapeAsStr(value.dtype, value.shape)}"
            )
    elif isinstance(value, bytes):
        try:
            return value.decode()
        except ValueError:
            return str(value)
    else:
        return str(value)


def typeAsStr(dtype: np.dtype) -> str:
    if h5py.check_string_dtype(dtype):
        return "string"
    elif h5py.check_vlen_dtype(dtype):
        return f"vlen array of {typeAsStr(h5py.check_vlen_dtype(dtype))}"
    elif h5py.check_enum_dtype(dtype):
        return "enum"
    elif dtype == h5py.ref_dtype:
        return "object reference"
    elif dtype == h5py.regionref_dtype:
        return "region reference"
    else:
        return str(dtype)


def typeAndShapeAsStr(dtype: np.dtype, shape: Tuple[int, ...]) -> str:
    if shape == ():
        return typeAsStr(dtype)
    else:
        return f"{typeAsStr(dtype)}{shape}"


def metadataFor(obj: Union[h5py.Group, h5py.Dataset]) -> Mapping[str, Any]:
    if isinstance(obj, h5py.Group):
        return {}
    else:
        return {
            "Shape": obj.shape,
            "DType": obj.dtype,
            "NBytes": obj.nbytes,
            "Max shape": obj.maxshape,
            "Chunk shape": obj.chunks,
            "Compression": obj.compression,
            "Compression opts": obj.compression_opts,
            "Filters": str(obj._filters),  # pylint: disable=protected-access
        }



# SPDX-FileCopyrightText: Copyright (c) 2024 Refeyn Ltd and other h5view contributors
# SPDX-License-Identifier: MIT

import importlib.metadata
import sys

from PySide6 import QtWidgets

from h5view.app import H5ViewWindow

__version__ = importlib.metadata.version("h5view")


def main() -> None:
    app = QtWidgets.QApplication(sys.argv)
    app.setOrganizationName("Refeyn Ltd")
    app.setApplicationName("h5view")
    app.setApplicationVersion(__version__)
    mw = H5ViewWindow()
    mw.show()
    if len(sys.argv) > 1:
        mw.open(sys.argv[1])
    app.exec()



# pylint: disable=all
# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'app.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QHBoxLayout, QHeaderView,
    QLabel, QMainWindow, QPlainTextEdit, QSizePolicy,
    QSpacerItem, QSplitter, QStatusBar, QTabWidget,
    QTableView, QToolBar, QToolButton, QTreeView,
    QVBoxLayout, QWidget)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(800, 600)
        self.actionOpen = QAction(MainWindow)
        self.actionOpen.setObjectName(u"actionOpen")
        self.actionShow_Image = QAction(MainWindow)
        self.actionShow_Image.setObjectName(u"actionShow_Image")
        self.actionShow_Image.setCheckable(True)
        self.actionAutoscale_Image = QAction(MainWindow)
        self.actionAutoscale_Image.setObjectName(u"actionAutoscale_Image")
        self.actionAutoscale_Image.setCheckable(True)
        self.actionAutoscale_Image.setChecked(True)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout_2 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.splitter = QSplitter(self.centralwidget)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Horizontal)
        self.treeView = QTreeView(self.splitter)
        self.treeView.setObjectName(u"treeView")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.treeView.sizePolicy().hasHeightForWidth())
        self.treeView.setSizePolicy(sizePolicy)
        self.splitter.addWidget(self.treeView)
        self.treeView.header().setVisible(False)
        self.rightPanels = QWidget(self.splitter)
        self.rightPanels.setObjectName(u"rightPanels")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(2)
        sizePolicy1.setVerticalStretch(1)
        sizePolicy1.setHeightForWidth(self.rightPanels.sizePolicy().hasHeightForWidth())
        self.rightPanels.setSizePolicy(sizePolicy1)
        self.verticalLayout_6 = QVBoxLayout(self.rightPanels)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.pathLabel = QLabel(self.rightPanels)
        self.pathLabel.setObjectName(u"pathLabel")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.pathLabel.sizePolicy().hasHeightForWidth())
        self.pathLabel.setSizePolicy(sizePolicy2)

        self.horizontalLayout_2.addWidget(self.pathLabel)

        self.copyPathButton = QToolButton(self.rightPanels)
        self.copyPathButton.setObjectName(u"copyPathButton")
        icon = QIcon(QIcon.fromTheme(u"edit-copy"))
        self.copyPathButton.setIcon(icon)

        self.horizontalLayout_2.addWidget(self.copyPathButton)

        self.typeLabel = QLabel(self.rightPanels)
        self.typeLabel.setObjectName(u"typeLabel")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.typeLabel.sizePolicy().hasHeightForWidth())
        self.typeLabel.setSizePolicy(sizePolicy3)
        self.typeLabel.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.horizontalLayout_2.addWidget(self.typeLabel)


        self.verticalLayout_6.addLayout(self.horizontalLayout_2)

        self.tabWidget = QTabWidget(self.rightPanels)
        self.tabWidget.setObjectName(u"tabWidget")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy4.setHorizontalStretch(2)
        sizePolicy4.setVerticalStretch(1)
        sizePolicy4.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy4)
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.verticalLayout_4 = QVBoxLayout(self.tab)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.spinContainer = QWidget(self.tab)
        self.spinContainer.setObjectName(u"spinContainer")
        sizePolicy5 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy5.setHorizontalStretch(0)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.spinContainer.sizePolicy().hasHeightForWidth())
        self.spinContainer.setSizePolicy(sizePolicy5)
        self.horizontalLayout = QHBoxLayout(self.spinContainer)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer = QSpacerItem(5, 5, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.verticalLayout_4.addWidget(self.spinContainer)

        self.viewWidgetContainer = QWidget(self.tab)
        self.viewWidgetContainer.setObjectName(u"viewWidgetContainer")
        sizePolicy6 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy6.setHorizontalStretch(0)
        sizePolicy6.setVerticalStretch(1)
        sizePolicy6.setHeightForWidth(self.viewWidgetContainer.sizePolicy().hasHeightForWidth())
        self.viewWidgetContainer.setSizePolicy(sizePolicy6)
        self.verticalLayout = QVBoxLayout(self.viewWidgetContainer)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.valueTextView = QPlainTextEdit(self.viewWidgetContainer)
        self.valueTextView.setObjectName(u"valueTextView")
        self.valueTextView.setReadOnly(True)

        self.verticalLayout.addWidget(self.valueTextView)

        self.valueTableView = QTableView(self.viewWidgetContainer)
        self.valueTableView.setObjectName(u"valueTableView")
        self.valueTableView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.valueTableView.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.valueTableView.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.valueTableView.horizontalHeader().setVisible(False)
        self.valueTableView.horizontalHeader().setStretchLastSection(True)
        self.valueTableView.verticalHeader().setVisible(False)

        self.verticalLayout.addWidget(self.valueTableView)


        self.verticalLayout_4.addWidget(self.viewWidgetContainer)

        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.verticalLayout_3 = QVBoxLayout(self.tab_2)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.attrTableView = QTableView(self.tab_2)
        self.attrTableView.setObjectName(u"attrTableView")
        self.attrTableView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.attrTableView.setTextElideMode(Qt.ElideNone)
        self.attrTableView.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.attrTableView.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.attrTableView.horizontalHeader().setVisible(False)
        self.attrTableView.horizontalHeader().setStretchLastSection(True)
        self.attrTableView.verticalHeader().setVisible(False)

        self.verticalLayout_3.addWidget(self.attrTableView)

        self.tabWidget.addTab(self.tab_2, "")
        self.tab_3 = QWidget()
        self.tab_3.setObjectName(u"tab_3")
        self.verticalLayout_5 = QVBoxLayout(self.tab_3)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.metadataTableView = QTableView(self.tab_3)
        self.metadataTableView.setObjectName(u"metadataTableView")
        self.metadataTableView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.metadataTableView.setTextElideMode(Qt.ElideNone)
        self.metadataTableView.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.metadataTableView.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.metadataTableView.horizontalHeader().setVisible(False)
        self.metadataTableView.horizontalHeader().setStretchLastSection(True)
        self.metadataTableView.verticalHeader().setVisible(False)

        self.verticalLayout_5.addWidget(self.metadataTableView)

        self.tabWidget.addTab(self.tab_3, "")

        self.verticalLayout_6.addWidget(self.tabWidget)

        self.splitter.addWidget(self.rightPanels)

        self.verticalLayout_2.addWidget(self.splitter)

        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.toolBar = QToolBar(MainWindow)
        self.toolBar.setObjectName(u"toolBar")
        self.toolBar.setMovable(False)
        self.toolBar.setFloatable(False)
        MainWindow.addToolBar(Qt.TopToolBarArea, self.toolBar)

        self.toolBar.addAction(self.actionOpen)
        self.toolBar.addAction(self.actionShow_Image)
        self.toolBar.addAction(self.actionAutoscale_Image)

        self.retranslateUi(MainWindow)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"H5View", None))
        self.actionOpen.setText(QCoreApplication.translate("MainWindow", u"Open", None))
        self.actionShow_Image.setText(QCoreApplication.translate("MainWindow", u"Show 2D as Image", None))
        self.actionAutoscale_Image.setText(QCoreApplication.translate("MainWindow", u"Autoscale Image", None))
        self.pathLabel.setText(QCoreApplication.translate("MainWindow", u"/x/y/z", None))
        self.copyPathButton.setText(QCoreApplication.translate("MainWindow", u"copy", None))
        self.typeLabel.setText(QCoreApplication.translate("MainWindow", u"float", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QCoreApplication.translate("MainWindow", u"Value", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QCoreApplication.translate("MainWindow", u"Attributes", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), QCoreApplication.translate("MainWindow", u"Metadata", None))
        self.toolBar.setWindowTitle(QCoreApplication.translate("MainWindow", u"toolBar", None))
    # retranslateUi




# SPDX-FileCopyrightText: Copyright (c) 2024 Refeyn Ltd and other h5view contributors
# SPDX-License-Identifier: MIT

import os
from typing import cast, List, Optional, Union

import h5py
import hdf5plugin  # pylint: disable=unused-import
from PySide6 import QtCore, QtGui, QtWidgets

import h5view
from h5view import gui, utils
from h5view.ui.app import Ui_MainWindow


class H5ViewWindow(Ui_MainWindow, QtWidgets.QMainWindow):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self._h5file: Optional[h5py.File] = None
        self._viewWidget: QtWidgets.QWidget = self.valueTableView
        self._regionSpins: List[QtWidgets.QSpinBox] = []
        self._settings = QtCore.QSettings()

        self.valueImageView = gui.PixmapWidget(self)
        self.valueImageView.setAutoscale(self.actionAutoscale_Image.isChecked())
        self.viewWidgetContainer.layout().addWidget(self.valueImageView)
        # Hide view widgets
        for i in range(self.viewWidgetContainer.layout().count()):
            widget = self.viewWidgetContainer.layout().itemAt(i).widget()
            if widget is not self.valueTableView:
                widget.hide()
        self.spinContainer.hide()

        self.pathLabel.setText("No file loaded")
        self.copyPathButton.hide()
        self.typeLabel.setText("")
        self.setWindowTitle(f"H5View {h5view.__version__}")

        self._connectSignals()

    def _connectSignals(self) -> None:
        self.actionOpen.triggered.connect(self._onOpen)
        self.actionShow_Image.triggered.connect(self._onUpdateShowIndex)
        self.actionAutoscale_Image.toggled.connect(self.valueImageView.setAutoscale)
        self.copyPathButton.clicked.connect(self._onCopyPathButtonClicked)

    def _onOpen(self) -> None:
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=self,
            dir=str(self._settings.value("openDir", ".")),
            filter="All files (*);;H5 files (*.h5 *.hdf5)",
        )
        if fname:
            self.open(fname)

    def _onShowIndex(self, idx: QtCore.QModelIndex) -> None:
        self._showIndex(idx)

    def _onUpdateShowIndex(self) -> None:
        if (
            self.treeView.selectionModel() is not None
            and len(self.treeView.selectionModel().selectedIndexes()) > 0
        ):
            self._showIndex(self.treeView.selectionModel().selectedIndexes()[0])

    def _onCopyPathButtonClicked(self) -> None:
        QtWidgets.QApplication.clipboard().setText(self.pathLabel.text())

    def open(self, fname: Union[str, os.PathLike]) -> None:
        self._settings.setValue("openDir", str(os.path.dirname(fname)))
        try:
            self._h5file = h5py.File(fname)
        except OSError as e:
            QtWidgets.QMessageBox.critical(self, "Could not open file", str(e))
        else:
            self._populateTreeView()
            self.setWindowTitle(
                f"H5View {h5view.__version__} - {os.path.abspath(fname)}"
            )

    def _populateTreeView(self) -> None:
        model = QtGui.QStandardItemModel()
        self._recursivePopulate(self._h5file, model.invisibleRootItem())
        self.treeView.setModel(model)
        self.treeView.selectionModel().select(
            model.index(0, 0), QtCore.QItemSelectionModel.SelectionFlag.SelectCurrent
        )
        self.treeView.selectionModel().currentChanged.connect(self._onShowIndex)
        self._showIndex(model.index(0, 0))

    def _recursivePopulate(
        self,
        group_or_dataset: Union[h5py.Group, h5py.Dataset],
        item: QtGui.QStandardItem,
    ) -> None:
        item.setData(group_or_dataset, cast(int, QtCore.Qt.ItemDataRole.UserRole))
        if isinstance(group_or_dataset, h5py.Group):
            for name, sub_grp_or_ds in group_or_dataset.items():
                subitem = QtGui.QStandardItem(name)
                subitem.setFlags(subitem.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)
                self._recursivePopulate(sub_grp_or_ds, subitem)
                item.appendRow(subitem)
        elif isinstance(group_or_dataset, h5py.Dataset):
            pass
        else:
            raise RuntimeError()

    def _showIndex(self, idx: QtCore.QModelIndex) -> None:
        group_or_dataset = idx.data(cast(int, QtCore.Qt.ItemDataRole.UserRole))
        self.pathLabel.setText(group_or_dataset.name)
        self.copyPathButton.show()

        attrModel = QtGui.QStandardItemModel()
        for i, (name, value) in enumerate(sorted(group_or_dataset.attrs.items())):
            attrModel.setItem(i, 0, QtGui.QStandardItem(name))
            attrModel.setItem(i, 1, QtGui.QStandardItem(utils.formatAsStr(value)))
        self.attrTableView.setModel(attrModel)
        self.attrTableView.resizeColumnsToContents()
        self.tabWidget.setTabText(1, f"Attributes ({len(group_or_dataset.attrs)})")

        metadataModel = QtGui.QStandardItemModel()
        for i, (name, value) in enumerate(
            sorted(utils.metadataFor(group_or_dataset).items())
        ):
            metadataModel.setItem(i, 0, QtGui.QStandardItem(name))
            metadataModel.setItem(i, 1, QtGui.QStandardItem(utils.formatAsStr(value)))
        self.metadataTableView.setModel(metadataModel)
        self.metadataTableView.resizeColumnsToContents()

        viewWidget = None
        regionSpins = 0
        if isinstance(group_or_dataset, h5py.Group):
            self.typeLabel.setText("group")
            viewWidget = self.valueTableView
            model = QtGui.QStandardItemModel()
            for i, (name, value) in enumerate(group_or_dataset.items()):
                model.setItem(i, 0, QtGui.QStandardItem(name))
                model.setItem(i, 1, QtGui.QStandardItem(utils.formatAsStr(value)))
            viewWidget.setModel(model)
            viewWidget.resizeColumnsToContents()

        elif isinstance(group_or_dataset, h5py.Dataset):
            self.typeLabel.setText(
                utils.typeAndShapeAsStr(group_or_dataset.dtype, group_or_dataset.shape)
            )
            if group_or_dataset.shape == ():
                viewWidget = self.valueTextView
                viewWidget.setPlainText(utils.formatAsStr(group_or_dataset))
            elif self.actionShow_Image.isChecked() and group_or_dataset.ndim >= 2:
                viewWidget = self.valueImageView
                self.valueImageView.setDataset(group_or_dataset)
                regionSpins = len(self.valueImageView.region())
            else:
                viewWidget = self.valueTableView
                viewWidget.setModel(gui.DatasetModel(group_or_dataset))
                regionSpins = len(viewWidget.model().region())
                dtypeWidth = gui.columnWidthFromDtype(
                    viewWidget, group_or_dataset.dtype
                )
                for i in range(viewWidget.model().columnCount()):
                    viewWidget.setColumnWidth(
                        i, dtypeWidth or gui.columnWidthRandomized(viewWidget, i)
                    )
        else:
            raise RuntimeError()

        if self._viewWidget:
            self._viewWidget.hide()

        if viewWidget:
            viewWidget.show()
            self._viewWidget = viewWidget

        while len(self._regionSpins) > regionSpins:
            self._regionSpins.pop()
            # Remove label and then spin
            self.spinContainer.layout().itemAt(
                len(self._regionSpins) * 2
            ).widget().setParent(None)
            self.spinContainer.layout().itemAt(
                len(self._regionSpins) * 2
            ).widget().setParent(None)

        while len(self._regionSpins) < regionSpins:
            sizePolicy = QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed
            )
            label = QtWidgets.QLabel(f"Dim {len(self._regionSpins)}")
            label.setSizePolicy(sizePolicy)
            self.spinContainer.layout().insertWidget(len(self._regionSpins) * 2, label)

            spin = QtWidgets.QSpinBox(self)
            spin.setSizePolicy(sizePolicy)
            spin.valueChanged.connect(self._onRegionChanged)
            self.spinContainer.layout().insertWidget(
                len(self._regionSpins) * 2 + 1, spin
            )
            self._regionSpins.append(spin)

        for i, spin in enumerate(self._regionSpins):
            spin.setRange(0, group_or_dataset.shape[i] - 1)

        # Remember the spacer!
        self.spinContainer.setVisible(self.spinContainer.layout().count() != 1)
        self._setRegion()

    def _onRegionChanged(self) -> None:
        self._setRegion()

    def _setRegion(self) -> None:
        region = tuple(s.value() for s in self._regionSpins)
        if self._viewWidget is self.valueTableView and isinstance(
            self.valueTableView.model(), gui.DatasetModel
        ):
            self.valueTableView.model().setRegion(region)
        elif self._viewWidget is self.valueImageView:
            self.valueImageView.setRegion(region)
